import logging


IPTOS_NORMAL = 0x0
IPTOS_LOWDELAY = 0x10
IPTOS_THROUGHPUT = 0x08
IPTOS_RELIABILITY = 0x04
IPTOS_MINCOST = 0x02
DEFAULT_LIMIT = 2 ** 20  # 1MB


log = logging.getLogger("sockio")


class ConnectionEOFError(ConnectionError):
    pass


class ConnectionTimeoutError(ConnectionError):
    pass



import asyncio
import functools
import threading
import urllib.parse

from . import aio


class BaseProxy:
    def __init__(self, ref):
        self._ref = ref

    def __getattr__(self, name):
        return getattr(self._ref, name)


def ensure_running(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if self.master and self.thread is None:
            self.start()
        return f(self, *args, **kwargs)

    return wrapper


class EventLoop:
    def __init__(self, loop=None):
        self.master = loop is None
        self.thread = None
        self.loop = loop
        self.proxies = {}

    def start(self):
        if self.thread:
            raise RuntimeError("event loop already started")
        if self.loop:
            raise RuntimeError("cannot run non master event loop")

        def run():
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
            started.set()
            self.loop.run_forever()

        started = threading.Event()
        self.thread = threading.Thread(name="AIOTH", target=run)
        self.thread.daemon = True
        self.thread.start()
        started.wait()

    def stop(self):
        if self.loop is None:
            if self.thread is None:
                raise RuntimeError("event loop not started")
        else:
            if self.thread is None:
                raise RuntimeError("cannot stop non master event loop")
        self.loop.call_soon_threadsafe(self.loop.stop)

    @ensure_running
    def run_coroutine(self, coro):
        return asyncio.run_coroutine_threadsafe(coro, self.loop)

    def _create_coroutine_threadsafe(self, corof, resolve_future):
        @functools.wraps(corof)
        def wrapper(obj, *args, **kwargs):
            coro = corof(obj._ref, *args, **kwargs)
            future = self.run_coroutine(coro)
            return future.result() if resolve_future else future

        return wrapper

    def _create_proxy_for(self, klass, resolve_futures=True):
        class Proxy(BaseProxy):
            pass

        for name in dir(klass):
            if name.startswith("_"):
                continue
            member = getattr(klass, name)
            if asyncio.iscoroutinefunction(member):
                member = self._create_coroutine_threadsafe(member, resolve_futures)
            setattr(Proxy, name, member)
        return Proxy

    @ensure_running
    def proxy(self, obj, resolve_futures=True):
        klass = type(obj)
        key = klass, resolve_futures
        Proxy = self.proxies.get(key)
        if not Proxy:
            Proxy = self._create_proxy_for(klass, resolve_futures)
            self.proxies[key] = Proxy
        return Proxy(obj)

    @ensure_running
    def tcp(self, host, port, resolve_futures=True, **kwargs):
        async def create():
            return aio.TCP(host, port, **kwargs)

        sock = self.run_coroutine(create()).result()
        return self.proxy(sock, resolve_futures)


DefaultEventLoop = EventLoop()
TCP = DefaultEventLoop.tcp


def socket_for_url(url, *args, **kwargs):
    addr = urllib.parse.urlparse(url)
    scheme = addr.scheme
    if scheme == "tcp":
        return TCP(addr.hostname, addr.port, *args, **kwargs)
    raise ValueError("unsupported sync scheme {!r} for {}".format(scheme, url))



__version__ = "0.14.0"


CONCURRENCY_MAP = {
    "sync": "sync",
    "syncio": "sync",
    "async": "async",
    "asyncio": "async",
}


def socket_for_url(url, *args, **kwargs):
    conc = kwargs.pop("concurrency", "async")
    concurrency = CONCURRENCY_MAP.get(conc)
    if concurrency == "async":
        from . import aio

        return aio.socket_for_url(url, *args, **kwargs)
    elif concurrency == "sync":
        from . import sio

        return sio.socket_for_url(url, *args, **kwargs)
    raise ValueError("unsupported concurrency {!r}".format(conc))



# -*- coding: utf-8 -*-

"""The setup script."""

import sys
from setuptools import setup, find_packages

with open("README.md") as f:
    description = f.read()

setup(
    name="sockio",
    author="Jose Tiago Macara Coutinho",
    author_email="coutinhotiago@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="Concurrency agnostic socket API",
    license="GPLv3+",
    long_description=description,
    long_description_content_type="text/markdown",
    keywords="socket, asyncio",
    packages=find_packages(include=["sockio"]),
    url="https://tiagocoutinho.github.io/sockio/",
    project_urls={
        "Documentation": "https://tiagocoutinho.github.io/sockio/",
        "Source": "https://github.com/tiagocoutinho/sockio/",
    },
    version="0.14.0",
    python_requires=">=2.7",
    zip_safe=True,
)
